# = rcxmigration.rb -- the main program of  migration tool
# 
# All Rights Reserved, Copyright (c) FUJITSU LIMITED 2010-2013
#
# $Id: rcxmigration.rb 1.28 09/05/13 12:04:35 FUJITSU LIMITED
#

# Determine platform environment (Windows or Linux OS)
#
require 'java'
require "#{RcxDir::manager_dir}/opt/FJSVrcxmr/lib/plugin/awwn.jar"
require 'syslog'

if java.lang.System.getProperty("os.name")  =~ /Win/   
  BASE_PATH = "../etc/rcx_migration_tool/"
  BASE_BIN_PATH = "../bin/"
  BASE_VAR_LOG_PATH = "../var/log/rcx_migration_tool/"
  BASE_CMD_STR_TMP = "#{BASE_BIN_PATH}rcxadm.exe"
  BASE_CMD_STR = "\"#{BASE_CMD_STR_TMP}\""
else  
  BASE_PATH = "/etc/opt/FJSVrcvmr/rcx_migration_tool/" 
  BASE_BIN_PATH="/opt/FJSVrcvmr/bin/"
  BASE_VAR_LOG_PATH = "/var/opt/FJSVrcvmr/log/rcx_migration_tool/"
  BASE_CMD_STR_TMP = "#{BASE_BIN_PATH}rcxadm"
  BASE_CMD_STR = "#{BASE_CMD_STR_TMP}"
end


#defination of file names
# configuration file
#
RCX_MIGRATION_CONF_FILE = BASE_PATH + "rcx_migration.conf"
VMGUEST_PRIORITY_CONF_FILE = BASE_PATH + "vmguest_priority.conf"
VMHOST_LIST_CONF_FILE = BASE_PATH + "vmhost_list.conf"
TRIGGER_EVENT_CONF_FILE = BASE_PATH + "trigger_event.conf"
VMHOST_GROUP_CONF_FILE = BASE_PATH + "vmhost_group.conf"
VMHOST_EXCLUDED_CONF_FILE = BASE_PATH + "vmguest_excluded.conf"
# log & lock file
INFO_FILE= BASE_VAR_LOG_PATH + "rcx_serverinfo."

#constant strings
# log
#
LOG_FILE_NUM = 3  # the number of log files is set to 3
LOG_FILE_SIZE = 10 * 1024 * 1024 #the file size of each log file is set to 10MB

#porgram control
#
POST_ACTION="POST_ACTION"
PRIOR_OFF_HOSTS = "PRIOR_OFF_HOSTS"  #default value, disable => false
DRS_USE = "DRS_USE"
MIGRATION_IN_MAINTENANCE_MODE = "MIGRATION_IN_MAINTENANCE_MODE"
VALUE_YES_STR = "yes"
VALUE_NO_STR = "no"
VALUE_NONE_STR = "none"
VALUE_POWEROFF_STR = "poweroff"
VALUE_SWITCHOVER_STR = "switchover"
VALUE_SWITCHOVER_AND_HOME_STR = "switchover_and_home"
VALUE_VMHOST_OPTION = "dest"
POST_SWITCHOVER_SLEEP = "post_switchover_sleep"
POST_SWITCHOVER_SLEEP_TIME = 80
VALUE_POST_ACTION_LIST = [
  VALUE_NONE_STR,
  VALUE_POWEROFF_STR,
  VALUE_SWITCHOVER_STR,
  VALUE_SWITCHOVER_AND_HOME_STR]
VALUE_PRIOR_OFF_HOSTS_LIST = [
  VALUE_YES_STR,
  VALUE_NO_STR]
VALUE_DRS_USE_LIST = [
  VALUE_YES_STR,
  VALUE_NO_STR]
VALUE_MIGRATION_IN_MAINTENANCE_MODE_LIST = [
  VALUE_YES_STR,
  VALUE_NO_STR]
DEFAULT_CONTROLLER = {
  POST_ACTION => VALUE_NONE_STR,
  PRIOR_OFF_HOSTS => VALUE_NO_STR,
  DRS_USE => VALUE_NO_STR,
  MIGRATION_IN_MAINTENANCE_MODE => VALUE_NO_STR}
WAITING_TIME_BEFORE_MIGRATION = 80
WAITING_TIME_BEFORE_POST_ACTION = 30

VALUE_PARALLEL_EXECUTION_NUM = 2
VALUE_MAX_RETRY_COUNT = 20

# command
#
CMD_STR_SERVER = " server"
VM_REGISTER_INFO = BASE_CMD_STR + CMD_STR_SERVER + " list -type all"
VM_MIGRATION = BASE_CMD_STR + CMD_STR_SERVER + " migrate"
SERVER_START = BASE_CMD_STR + CMD_STR_SERVER + " start"
SERVER_STOP = BASE_CMD_STR + CMD_STR_SERVER + " stop"
SWITCHOVER = BASE_CMD_STR + CMD_STR_SERVER + " switchover"
VM_HOME = BASE_CMD_STR + CMD_STR_SERVER + " migrate -home"
SET_VM_MAINTENANCE = BASE_CMD_STR + CMD_STR_SERVER + " set"

#special chars
#
EQUAL_CHAR = "="
NULL_STR = ""
SPACE_CHAR = " "
COLON_CHAR = ":"

#definations of serveral conditions for registered servers 
#
MAINTENANCE_ON = "ON"  #in the case of getting the server status with a CLI command
STATUS_STOP = "stop"
STATUS_NORMAL = "normal"
STATUS_ERROR = "error"
STATUS_UNKNOWN = "unknown"
TYPE_VMHOST = "vm_host"
TYPE_VMGUEST = "vm_guest"

# unsupported error
UNSUPPORTED_ERR="FJSVrcx:ERROR:67280"


# === ResourceData Class
# Resource data exists by server list command.
class ResourceData
  attr_accessor :name, :vmhost, :admin_ip, :vmguests, :status
  def initialize(data)
    @server_name = data[0]
    @name = data[1]
    @type = data[2]
    @vmhost = data[3]
    @admin_ip = data[4]
    @status = data[5]
    @maintenance = data[6]
    @vmguests = []
  end

  # = check resource type
  def type?(type)
    return @type == type
  end

  # = check maintenance mode
  def maintenance?
    return @maintenance == MAINTENANCE_ON
  end

  # = check status
  def status?(status)
    return @status == status
  end

  # ===Methods: poweroff_executable?
  # it is used to shutdown the power of the source vmhost 
  # ====Parameters & return value
  # *Return*: [Bool]  
  #
  def poweroff_executable?
    @vmguests.each { |data|
      unless data.status?(STATUS_STOP)
        RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] poweroff_executable?(): server #{name} cannot be stopped!")
        return false
      end
    }
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] poweroff_executable?(): server #{name} can be stopped!")
    return true
  end

  # == on_vmguests?
  # check if there are vmguests on vmhost
  # == parameters
  #
  def has_active_vmguests?
    result = false
    @vmguests.each do |vmg|
      unless vmg.status?(STATUS_STOP)
        result = true
        break
      end
    end
    return result
  end
end

# == ServerList class
# it can be used for getting information of source vm hosts, vm guests, and dest vm hosts.
# Also, it includes checking out the power status of dest vmhost and source vnhost.
# ===Parameters
# @rsc_list:[Array] it is used for keeping the information of registered vm hosts and vm gusets
#
class ServerList
  # ===Methods: initialize
  # it is used for create a array named @rsc_list
  #
  def initialize(ip)
    @rsc_list = []
    @src_ipaddr = ip
  end
  
  def exec_cmd(cmd,thread_id=0)
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] exec_cmd() -- start cmd=#{cmd} thread_id=#{thread_id} ")
    infofile = "#{INFO_FILE}#{@src_ipaddr.gsub(".", "_")}_#{thread_id}.tmp"
    result = true
    File.open(infofile, "w") do |openinfofile|
      $stdout = openinfofile
      result = false unless system(cmd)
      $stdout = STDOUT
    end
    unless File.exist?(infofile) 
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] exec_cmd(): file not found: #{infofile}")
      return false
    end
    File.open(infofile) do |file|
      file.each_line do |line|
        if cmd.include?(SET_VM_MAINTENANCE) && line.include?(UNSUPPORTED_ERR)
          result = true
          RcxLogger.logger.info("[rcx_migration_tool] skip unsupported error. cmd=#{cmd}")
        end
        RcxLogger.logger.info("[rcx_migration_tool] #{line.strip}")
      end
    end
    File.delete(infofile)
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] exec_cmd() -- end  cmd=#{cmd} thread_id=#{thread_id}")
    return result
  end

  def exec_cmd_thread(cmd_path,thread_id=0)
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] exec_cmd_thread() -- start cmd_path=#{cmd_path} thread_id=#{thread_id} ")
    result = true
    success_code = 0
    rmc_error = -1
   
    cmd = cmd_path.split
    
    begin 
      rmc_facade = com.fujitsu.rcx.vio.awwn.rmc.ProcessExecRMCFacade.new 
       
      # convert string array ruby => java(java.lang.String[]) 
      # call com.fujitsu.rcx.vio.awwn.rmc.ProcessExecRMCFacade#execute 
      command_result = rmc_facade.execute(cmd.to_java(:string)) 
      # get result from com.fujitsu.rcx.vio.awwn.rmc.CommandResult 
      return_code    = command_result.getErrorCode 
      output         = command_result.getMessags[0] 
      stderr_message = command_result.getMessags[1] 
      
    rescue Exception => e 
      output = e.message 
      return_code = rmc_error
      exit_code = 3
      stderr_message = e.message 
    end 

    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] exec_cmd_thread()  return_code=#{return_code}  output=#{output} stderr_message=#{stderr_message} thread_id=#{thread_id} ")

    if return_code != success_code
      result = false
    end

    output.each_line do |line|
      if cmd_path.include?(SET_VM_MAINTENANCE) && line.include?(UNSUPPORTED_ERR)
        result = true
        RcxLogger.logger.info("[rcx_migration_tool] skip unsupported error. cmd_path=#{cmd_path}")
      end
      RcxLogger.logger.info("[rcx_migration_tool] #{line.strip}")
    end
    
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] exec_cmd_thread() -- end  cmd_path=#{cmd_path} thread_id=#{thread_id}")
    return result
  end

  # ===Methods: update_rsc_list
  # it is used for getting all the registration information
  # ====Parameters & return value
  # *Return*: [Bool] 
  #             true: getting the registration information successfully    
  #             false: failed to getting the registration information
  #
  def update_rsc_list
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] update_rsc_list() -- start")
    infofile = "#{INFO_FILE}#{@src_ipaddr.gsub(".", "_")}.tmp"
    @rsc_list = []
    vmguests = {}
    File.open(infofile, "w") do |openinfofile|
      $stdout = openinfofile
      unless system(VM_REGISTER_INFO)
        RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] update_rsc_list(): #{VM_REGISTER_INFO}" + " exec error (" + $?.to_s + ")")
        return false
      end
      $stdout = STDOUT
    end

    unless File.exist?(infofile) 
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] update_rsc_list(): file not found: #{infofile}")
      return false
    end

    RcxLogger.logger.info("[rcx_migration_tool] --")
    begin
      File.open(infofile) do |file|
        file.each_line do |line|
          RcxLogger.logger.info("[rcx_migration_tool] #{line.strip}")
          line.tr!("\t", SPACE_CHAR)
          raw_data = line.split(SPACE_CHAR)
          data = []
          data << raw_data[0]
          sv_name = ""
          (1..raw_data.size-6).each do |i|
            sv_name += raw_data[i]
            sv_name += "\s+" if i != raw_data.size-6
          end
          regexp_sv_name = Regexp.new(sv_name)
          data << line.scan(regexp_sv_name)[0]
          (raw_data.size-5..raw_data.size-1).each do |i|
            data << raw_data[i]
          end
          rsc_data = ResourceData.new(data)
          if rsc_data.type?(TYPE_VMGUEST)
            vmguests[rsc_data.vmhost] = [] unless vmguests[rsc_data.vmhost]
            vmguests[rsc_data.vmhost] << rsc_data
          end
          @rsc_list << rsc_data
        end
      end
      File.delete(infofile)
      vmguests.each_key do |vmh_name|
        vmh = @rsc_list.find { |rsc| rsc.name == vmh_name}
        vmh.vmguests = vmguests[vmh_name] if vmh
      end
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] update_rsc_list(): detail=#{ex.class}/#{ex.message}")
      return false
    end

    #RcxLogger.logger.info("[rcx_migration_tool] -- ")

    @rsc_list.delete_at(0)
    @rsc_list.delete_at(0)        
    RcxLogger.logger.info("[rcx_migration_tool] update_rsc_list() -- end ")
    return true
  end

  # ===Methods: get_src_vmhost
  # it is used for getting information of source vm hosts
  # ====Parameters & return value
  # *Return*: [String]  
  #              vmhost: the vm_host name of the trap server, where happened a trap
  #              NULL_STR: failed to getting the vm_host name
  #
  def get_src_vmhost
    begin
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_src_vmhost() -- start")
      vmhost = nil
      @rsc_list.each { |data|
        if data.type?(TYPE_VMHOST) && data.admin_ip == @src_ipaddr
          vmhost = data
          RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_src_vmhost(): source VMhost => ip=#{@src_ipaddr}, name=#{vmhost.name}]")
          break
        end
      }
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_src_vmhost(): source VMhost not found.") unless vmhost
      return vmhost
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] get_src_vmhost(): detail=#{ex.class}/#{ex.message}")
      return nil
    end
  end


  # ===Methods: get_status
  # it is used for getting information of object power status
  # ====Parameters & return value
  # *Return*: [String]  
  #              object power status
  #              NULL_STR: failed to getting the power status
  #
  def get_status(object_name)
    begin
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_status() -- start Resource data=#{object_name}")
      @rsc_list.each { |data|
        if data.name == object_name
          RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_status()  Resource data=#{object_name} status=#{data.status}")
          return data.status
        end
      }
      
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_status(): error Resource data=#{object_name} .")
      return nil
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] get_status(): detail=#{ex.class}/#{ex.message}")
      return nil
    end
  end


  # ===Methods: get_mv_vmguests
  # it is used for getting information of vm guests 
  # ====Parameters & return value
  # _src_vmhost_: [String] a vm_host name 
  # *Return*: [Array]  
  #              vmguests: the vmguests information at the trap server, where happened a trap
  #              []: there isn't any vm_guest information
  #
  def get_mv_vmguests(src_vmhost)
    begin
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_mv_vmguests() -- start")
      vmguests = []
      rsc_list_vmg = src_vmhost.vmguests
      rsc_list_vmg.sort! {|a,b| a.name <=> b.name}
    
    # read excluded vmguests from vmguest_excluded.conf
    vmguests_excluded_conf = []
    if File.exist?(VMHOST_EXCLUDED_CONF_FILE)
        File.open(VMHOST_EXCLUDED_CONF_FILE) do |file|
          file.each_line do |line|
            vmguests_excluded_conf << line.strip
          end
        end
      else
        RcxLogger.logger.warn("[rcx_migration_tool] [#{self.class}] get_mv_vmguests(): #{VMHOST_EXCLUDED_CONF_FILE} is not exist!")
      end
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_mv_vmguests(): vmguests_excluded=#{vmguests_excluded_conf.join(",")}")
    
      if File.exist?(VMGUEST_PRIORITY_CONF_FILE) 
        File.open(VMGUEST_PRIORITY_CONF_FILE) do |file|
          file.each_line do |line|
            guest_name_conf = line.strip
            data = rsc_list_vmg.find {|vmg| vmg.name == guest_name_conf}
            if data && !data.status?(STATUS_STOP)
              vmguests << data
            end
          end
        end      
      end
      rsc_list_vmg.each do |data|
        if !data.status?(STATUS_STOP) && !vmguests.include?(data) && !vmguests_excluded_conf.include?(data.name)
          vmguests << data
        end
      end
      vmguests_name = vmguests.collect {|vmg| vmg.name }
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_mv_vmguests(): mv_vmguest is [#{vmguests_name.join(",")}] on [#{src_vmhost.name}]")
      return vmguests
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] get_mv_vmguests(): detail=#{ex.class}/#{ex.message}")
      return []
    end
  end
 
  # ===Methods: get_dest_vmhosts
  # it is used for getting information of dest vm hosts
  # ====Parameters & return value
  # _src_vmhost_: [String] a vm_host name 
  # _group_vmhosts_name_: [ARRAY] vmhosts list
  # _prior_off_hosts_: [String] PRIOR_OFF_HOSTS controller
  # _specified_dest_: [ARRAY] vmhosts list (dest only)
  # *Return*: [Array]  
  #             dest vmhost: the dest vm_host name list of the trap server, where happened a trap
  #
  def get_dest_vmhosts(src_vmhost, group_vmhosts_name, prior_off_hosts,specified_dest,migration_in_maintenance_mode)
    begin
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_dest_vmhosts(): -- start")
    # if src_vmhost is specified destination, destination is not set for the src_vmhost.
    return [] if specified_dest.include?(src_vmhost.name)
      dest_vmhosts = []
      stop_vmhosts = []
      group_vmhosts_name.delete(src_vmhost.name)
    
    group_vmhosts_name = specified_dest if specified_dest.size > 0
      group_vmhosts_name.each do |vmhost_name|
        data = @rsc_list.find {|vmh| vmhost_name == vmh.name}
        if data && !data.status?(STATUS_ERROR) && !data.status?(STATUS_UNKNOWN)
      if migration_in_maintenance_mode == VALUE_YES_STR || (migration_in_maintenance_mode == VALUE_NO_STR && !data.maintenance?)
            dest_vmhosts << data
            stop_vmhosts << data if data.status?(STATUS_STOP)
        end
        end
      end

      if dest_vmhosts.size > 0 && prior_off_hosts == VALUE_YES_STR
        dest_vmhosts.delete_if {|vmh| stop_vmhosts.include?(vmh) }
        dest_vmhosts = stop_vmhosts + dest_vmhosts
      end
      dest_vmhosts_name = dest_vmhosts.collect {|vmh| vmh.name }
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_dest_vmhosts(): dest_vmhosts=#{dest_vmhosts_name.join(",")}")
      return dest_vmhosts
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] get_dest_vmhosts(): detail=#{ex.class}/#{ex.message}")
      return []
    end
  end
end

class ConfReader
  def initialize
    @controller = DEFAULT_CONTROLLER
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}]:controller default=#{@controller.inspect}")
  end

  # === get_controller_value
  # get configured value from conf file
  # == parameters
  # _parameter_: configured parameter in conf file
  # _group_name: vmhost group name
  # == return
  #   value: configured value of parameter
  def get_controller_value(parameter, group_name = nil)
    begin
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_controller_value() -- start")
      key, value = nil, nil
      if group_name
        key = "#{parameter}-#{group_name}"
        value = @controller[key]
      end
      unless value
        key = parameter
        value = @controller[key]
      end
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_controller_value() #{key}=#{value}")
      return value
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] get_controller_value(): detail=#{ex.class}/#{ex.message}")
      return nil
    end
  end

  # = read_rcx_migration_conf
  # read configration for the main program from rcxmigration.conf
  # ==== Parameters
  # *Return*:
  def read_rcx_migration_conf
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] read_rcx_migration_conf() -- start")
    #read values from rcx_migration.conf
    unless File.exist?(RCX_MIGRATION_CONF_FILE)
      RcxLogger.logger.warn("[rcx_migration_tool] #{RCX_MIGRATION_CONF_FILE} is not exist!")
      return
    end
    begin
      File.open(RCX_MIGRATION_CONF_FILE) do |file|
        file.each_line do |line|    
          line.gsub!(/#.*$/, NULL_STR)                              # delete comment
          data = line.split(EQUAL_CHAR)
          if data[0] && data[1]
            data[0].strip!
            data[1].strip!
            key = data[0].gsub(/-.*$/, NULL_STR)
            value_list = eval "VALUE_#{key}_LIST"
            if value_list.include?(data[1])
              @controller[data[0]] = data[1]
            else
              RcxLogger.logger.warn("[rcx_migration_tool] #{data[0]}: value is invalid!" )
            end
          end
        end
      end
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] read_rcx_migration_conf():controller current=#{@controller.inspect}")
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] read_rcx_migration_conf(): detail=#{ex.class}/#{ex.message}")
    end
  end

  # = trigger_event_conf_include?
  # check whether a event is on the trigger_event or not.
  # ==== Parameters
  # message:[String] received trap message
  # *Return*: 
  # [Boolean] event is on the trigger_event.conf or not 
  #
  def trigger_event_conf_include?(message)
    RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] trigger_event_conf_include?() -- start")
    unless File.exist?(TRIGGER_EVENT_CONF_FILE)
      RcxLogger.logger.warn("[rcx_migration_tool] #{TRIGGER_EVENT_CONF_FILE} is not exist!" )
      return false
    end
    begin
      File.open(TRIGGER_EVENT_CONF_FILE) do |file|
        file.each_line do |line|    
          line.gsub!(/#.*$/, NULL_STR)                              # delete comment
          data = line.split(EQUAL_CHAR)
          if data[0] && data[1]
            cond_str1_tmp = data[0].split(COLON_CHAR)[1]
            cond_str1 = cond_str1_tmp.strip
            cond_str2 = data[1].strip
            if message.include?(cond_str1) && cond_str2 == "true"
              RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] trigger_event_conf_include?(): [#{cond_str1}] existed!")
              return true
            end
          end
        end 
      end  
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] trigger_event_conf_include?(): detail=#{ex.class}/#{ex.message}")
      return false
    end
    return false
  end

  # === get_vmhost_group
  #get vmhosts in range of src_vmhost
  # == parameters
  # _src_vmhost_: [String] source vmhost name
  # *return* 
  # group_name: [String] vmhost group name
  # vmhosts: [Array] vmhost list in the same group
  def get_vmhost_group(src_vmhost)
    begin
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_vmhost_group() -- start")
      group_name, vmhosts = nil, []
      tmp_group_name, tmp_vmhosts,  tmp_specified_dest = nil, {}, {}
    vmhost_and_option, specified_dest = nil, []
      if File.exist?(VMHOST_GROUP_CONF_FILE)
        File.open(VMHOST_GROUP_CONF_FILE) do |file|
          while line = file.gets
            line.gsub!(/#.*$/, NULL_STR)
            if line =~ /\[.*\]/
              tmp_group_name = line.scan(/\[(.*)\]/)[0][0]
              tmp_vmhosts[tmp_group_name] = []
        tmp_specified_dest[tmp_group_name] = []
            elsif line && line.strip != NULL_STR
        vmhost_and_option = line.strip.split(/\s*,\s*/)
        
        # skip line if two options or more in the line
        #  vmhost_and_option[0]:vmhost
        #  vmhost_and_option[1]:option
        if vmhost_and_option.size < 3 && (vmhost_and_option[1] == nil || vmhost_and_option[1] == VALUE_VMHOST_OPTION)  && tmp_group_name
                tmp_vmhosts[tmp_group_name] << vmhost_and_option[0]
          tmp_specified_dest[tmp_group_name] << vmhost_and_option[0] if vmhost_and_option[1]
              end
      end
          end
        end
      else
        RcxLogger.logger.warn("[rcx_migration_tool] [#{self.class}] get_vmhost_group(): #{VMHOST_GROUP_CONF_FILE} is not exist!")
      end
      tmp_vmhosts.each_key do |key|
        if tmp_vmhosts[key].include?(src_vmhost.name)
          group_name, vmhosts, specified_dest = key, tmp_vmhosts[key], tmp_specified_dest[key]
          break
        end
      end
      RcxLogger.logger.info("[rcx_migration_tool] [#{self.class}] get_vmhost_group(): group_name=#{group_name}, VMHosts=#{vmhosts.join(",")}, specified_dest=#{specified_dest.join(",")}")
      return group_name, vmhosts, specified_dest
    rescue Exception => ex
      RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] get_vmhost_group(): detail=#{ex.class}/#{ex.message}")
      return nil, [], []
    end
  end
end

# === exec_os_cmd
# excute os command
# == parameters
# _cmd_: [String] excute os command
# *return*  none
def exec_os_cmd(cmd)
  begin
    system(cmd)
  rescue
  RcxLogger.logger.error("[rcx_migration_tool] [#{self.class}] exec_os_cmd(): execute #{cmd} error")
  end
end

#
#                                                       
# == main program of rcx_migration.rb   
#                                                       
#                                                       

# create a folder for log files
#
begin
  unless File.exist?(BASE_VAR_LOG_PATH)
    Dir.mkdir(BASE_VAR_LOG_PATH)
  end
rescue
  exit 1
end

# = check argument
# ==== Parameters
# message:: [String] a string of a trap message
# ip_addr:: [String] an ip address where a trap has happened
#

if java.lang.System.getProperty("os.name")  =~ /Win/   
  message = ARGV[2]
  ip_addr = ARGV[3]
else
  message = ARGV[3]
  ip_addr = ARGV[4]
end

if message == nil || ip_addr == nil || message == NULL_STR || ip_addr == NULL_STR 
  RcxLogger.logger.error("[rcx_migration_tool] ARGV is empty.")
  exit 1
end

lockfile = "#{BASE_VAR_LOG_PATH}rcx_migration.#{ip_addr.gsub(".", "_")}.lock"

RcxLogger.logger.info("[rcx_migration_tool] ========rcx_migratioin program: start=======")
RcxLogger.logger.info("[rcx_migration_tool] start: rcxmigration.rb")
RcxLogger.logger.info("[rcx_migration_tool] ip_address: [#{ip_addr}], message: [#{message}]")

# = file lock
# Using the lock file, only one trap will be processing among traps from a same IP 
# ==== Parameters
# lockfile: [String] the file path of a lockfile to a ip 
#
lock_fd = File.open(lockfile, "w")
exit_code = 0
begin
  if lock_fd.flock(File::LOCK_EX | File::LOCK_NB) == false 
    exit_code = 2
    raise StandardError.new("already running")
  end

  # read config ingotmation from rcx_migration.conf
  conf_reader = ConfReader.new
  conf_reader.read_rcx_migration_conf

  # compare the trap message with the events in the trigger_event.conf
  unless conf_reader.trigger_event_conf_include?(message)
    exit_code = 0
    raise StandardError.new("event not found. [#{message}]")
  end
  RcxLogger.logger.info("[rcx_migration_tool] event found. [#{message}]")

  # get the registered servers information 
  server_list = ServerList.new(ip_addr)
  unless server_list.update_rsc_list
    exit_code = 3
    raise StandardError.new("failed to get registered server list")
  end

  # = get src vmhost
  src_vmhost = server_list.get_src_vmhost
  if src_vmhost == nil 
    exit_code = 0
    raise StandardError.new("#{ip_addr} is not VMhost.")
  end

  # = get group name, vmhosts and destination hosts which src_vmhost is in.
  group_name, group_vmhosts_name, specified_dest = conf_reader.get_vmhost_group(src_vmhost)

  # = get drs_use, prior_off_hosts, post_action from conf file
  drs_use = conf_reader.get_controller_value(DRS_USE, group_name)
  prior_off_hosts = conf_reader.get_controller_value(PRIOR_OFF_HOSTS, group_name)
  post_action = conf_reader.get_controller_value(POST_ACTION, group_name)
  migration_in_maintenance_mode = conf_reader.get_controller_value(MIGRATION_IN_MAINTENANCE_MODE, group_name)

  # = check status of src vmhost(maintenance mode)
  if migration_in_maintenance_mode == VALUE_NO_STR && src_vmhost.maintenance?
    exit_code = 0
    raise StandardError.new("#{ip_addr} is in maintenance mode. MIGRATION_IN_MAINTENANCE_MODE=#{migration_in_maintenance_mode}")
  end

  if drs_use == VALUE_NO_STR
    # = get mv vmguest
    mv_vmguests = server_list.get_mv_vmguests(src_vmhost)

    if mv_vmguests.size > 0
      # = get dest vmhost
      dest_vmhosts = server_list.get_dest_vmhosts(src_vmhost, group_vmhosts_name, prior_off_hosts,specified_dest,migration_in_maintenance_mode)
      if dest_vmhosts.size == 0
        exit_code = 4
        raise StandardError.new("dest VMhost not found.")
      end

      threads = []
      running_thread_count = 0
      parallel_execution_num = VALUE_PARALLEL_EXECUTION_NUM

      if java.lang.System.getProperty("os.name")  =~ /Win/   
        # write log to application log    
        os_cmd = "eventcreate /t information /id 211 /d \"FJSVrcx:INFO:21168:#{src_vmhost.name}:rcx_migration_tool execution started\" /l application /so JSE_SWRC_FJSVRCX"
      exec_os_cmd(os_cmd)
      else
        Syslog.open(ident="rcx", facility=Syslog::LOG_USER) {|syslog|
	        syslog.log(Syslog::LOG_INFO, "%s", "FJSVrcx:INFO:21168:#{src_vmhost.name}:rcx_migration_tool execution started")
	      }
      end
      mv_vmguests.each_with_index do |mv_vmguest, id|
        #count = id + 1
        
        dest_vmhost_id = id % dest_vmhosts.size
        dest_vmhost = dest_vmhosts[dest_vmhost_id]
        
        # poweron dest_vmhost
        RcxLogger.logger.info("[rcx_migration_tool] status of #{dest_vmhost.name} is #{dest_vmhost.status}.")
        if dest_vmhost.status?(STATUS_STOP)
          tmp_cmd = SERVER_START+ " -name " + dest_vmhost.name
          RcxLogger.logger.info("[rcx_migration_tool] poweron start #{dest_vmhost.name}")
          unless server_list.exec_cmd(tmp_cmd) 
            exit_code = 3
            raise StandardError.new("poweron execute #{tmp_cmd} error.")
          else
            RcxLogger.logger.info("[rcx_migration_tool] poweron executed #{tmp_cmd}")
            # waiting for WAITING_TIME_BEFORE_MIGRATION before migration
            # Migration cannot be executed soon after poweron.
            RcxLogger.logger.info("[rcx_migration_tool] sleep start before migration. time=#{WAITING_TIME_BEFORE_MIGRATION}")
            sleep(WAITING_TIME_BEFORE_MIGRATION)
            
            unless server_list.update_rsc_list
              exit_code = 3
              raise StandardError.new("failed to get registered server list")
            end
            
            dest_vmhost.status = server_list.get_status(dest_vmhost.name)
            if dest_vmhost.status == nil 
              exit_code = 3
              raise StandardError.new("status of #{dest_vmhost.name} is null.")
            end
          end
        end

        tmp_cmd = "#{VM_MIGRATION} -name \"#{mv_vmguest.name}\" -to #{dest_vmhost.name}"
          
        #two vmguests will be migrated to dest vmhost each time
        #e.g. count%2
        #
        #if ((count % 2 == 1) && (count != mv_vmguests.size)) 
        #  tmp_cmd += " -nowait"
        #end
        
        thread_id = id+1
        running_thread_count +=1
        threads[thread_id] = Thread.new do
          RcxLogger.logger.info("[rcx_migration_tool] exec migration src=#{src_vmhost.name} dest=#{dest_vmhost.name} mv=#{mv_vmguest.name} -- start")
          unless server_list.exec_cmd_thread(tmp_cmd,thread_id)
            exit_code = 3
            RcxLogger.logger.error("[rcx_migration_tool] exec_cmd error migration src=#{src_vmhost.name} dest=#{dest_vmhost.name} mv=#{mv_vmguest.name}  thread_id=#{thread_id}")
            raise StandardError.new("execute #{tmp_cmd} thread_error.")
          else
            RcxLogger.logger.info("[rcx_migration_tool] thread_execute #{tmp_cmd} thread_id=#{thread_id}")
          end
        end
        
        if running_thread_count%parallel_execution_num == 0 || thread_id == mv_vmguests.size
          for i in (thread_id-running_thread_count+1)..thread_id
            threads[i].join
          end
          running_thread_count=0
        end
      end
      if java.lang.System.getProperty("os.name")  =~ /Win/   
        # write completed log to application log
        os_cmd = "eventcreate /t information /id 211 /d \"FJSVrcx:INFO:21169:#{src_vmhost.name}:rcx_migration_tool execution completed\" /l application /so JSE_SWRC_FJSVRCX"
        exec_os_cmd(os_cmd)
      else
        Syslog.open(ident="rcx", facility=Syslog::LOG_USER) {|syslog|
	        syslog.log(Syslog::LOG_INFO, "%s", "FJSVrcx:INFO:21169:#{src_vmhost.name}:rcx_migration_tool execution completed")
	      }
      end
    else
      RcxLogger.logger.warn("[rcx_migration_tool] mv_vmguest not found.")
    end
  end

  server_list.update_rsc_list
  src_vmhost = server_list.get_src_vmhost
  if (migration_in_maintenance_mode == VALUE_NO_STR && src_vmhost.maintenance?) || src_vmhost == nil 
    exit_code = 0
    raise StandardError.new("#{ip_addr} is not VMhost or in maintenance mode. MIGRATION_IN_MAINTENANCE_MODE=#{migration_in_maintenance_mode}")
  end

# set vm maintenance
  tmp_cmd = "#{SET_VM_MAINTENANCE} -name #{src_vmhost.name} -attr vmm_mode=maintenance"
  if drs_use == VALUE_YES_STR || !src_vmhost.has_active_vmguests?
    unless server_list.exec_cmd(tmp_cmd) 
      exit_code = 3
      raise StandardError.new("execute #{tmp_cmd} error.")
    end
  end

  # = execute post_action
  loop_count = 0
  max_retry_count = VALUE_MAX_RETRY_COUNT
  
  while loop_count < max_retry_count
    if server_list.update_rsc_list
      break
    else
      loop_count += 1
      if loop_count == max_retry_count
        exit_code = 3
        raise StandardError.new("failed to get registered server list")
      end
      sleep(WAITING_TIME_BEFORE_POST_ACTION)
      RcxLogger.logger.info("[rcx_migration_tool] Retry #{tmp_cmd} Retry Count=#{loop_count}")
    end
  end

  src_vmhost = server_list.get_src_vmhost
  if (migration_in_maintenance_mode == VALUE_NO_STR && src_vmhost.maintenance?) || src_vmhost == nil 
    exit_code = 0
    raise StandardError.new("#{ip_addr} is not VMhost or in maintenance mode. MIGRATION_IN_MAINTENANCE_MODE=#{migration_in_maintenance_mode}")
  end
  
  tmp_cmds = []
  unless src_vmhost.has_active_vmguests?
    case post_action
    when VALUE_POWEROFF_STR
      if src_vmhost.poweroff_executable?
        tmp_cmds << "#{SERVER_STOP} -name #{src_vmhost.name} -nowait"
      end
    when VALUE_SWITCHOVER_STR
      tmp_cmds << "#{SWITCHOVER} -name #{src_vmhost.name}"
      tmp_cmds << POST_SWITCHOVER_SLEEP
      tmp_cmds << "#{SET_VM_MAINTENANCE} -name #{src_vmhost.name} -attr vmm_mode=active"
    when VALUE_SWITCHOVER_AND_HOME_STR
      tmp_cmds << "#{SWITCHOVER} -name #{src_vmhost.name}"
      tmp_cmds << POST_SWITCHOVER_SLEEP
      tmp_cmds << "#{SET_VM_MAINTENANCE} -name #{src_vmhost.name} -attr vmm_mode=active"
      tmp_cmds << "#{VM_HOME} -nowait" if drs_use == VALUE_NO_STR
    else
      # no action
    end
  else
    RcxLogger.logger.info("[rcx_migration_tool] #{src_vmhost.name} has active vmguests, so post_action cannot be executed.")
  end
  RcxLogger.logger.info("[rcx_migration_tool] post_action=#{tmp_cmds.inspect}")
  tmp_cmds.each do |tmp_cmd|
    if tmp_cmd == POST_SWITCHOVER_SLEEP
      RcxLogger.logger.info("[rcx_migration_tool] sleep start. time=#{POST_SWITCHOVER_SLEEP_TIME}")
      sleep(POST_SWITCHOVER_SLEEP_TIME)
      RcxLogger.logger.info("[rcx_migration_tool] sleep end.")
    else
      unless server_list.exec_cmd(tmp_cmd) 
        exit_code = 3
        raise StandardError.new("execute #{tmp_cmd} error.")
      end
    end
  end
    
rescue Exception => ex
  RcxLogger.logger.error("[rcx_migration_tool] error occured. detail=#{ex.class}/#{ex.message}. backtrace=\n#{ex.backtrace.join("\n")}" )
  if java.lang.System.getProperty("os.name")  =~ /Win/   
    # write completed log to application log
    os_cmd = "eventcreate /t error /id 611 /d \"FJSVrcx:ERROR:61143:rcx_migration_tool:failed\" /l application /so JSE_SWRC_FJSVRCX"
    exec_os_cmd(os_cmd)
  else
    Syslog.open(ident="rcx", facility=Syslog::LOG_USER) {|syslog|
	        syslog.log(Syslog::LOG_ERR, "%s", "FJSVrcx:ERROR:61143:rcx_migration_tool:failed")
	  }
  end
ensure
  lock_fd.flock(File::LOCK_UN)
  lock_fd.close
  File.delete(lockfile)
  RcxLogger.logger.info("[rcx_migration_tool] end: rcx_migration.rb #{ip_addr} #{message}")
  RcxLogger.logger.info("[rcx_migration_tool] ========rcx_migratioin program: end=======")
  exit exit_code
end
